import csv
import os
import random
from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

import numpy as np
import pandas as pd
import PIL.Image as Image
import torch
import torch.utils
import torchvision

from ..helper import helper
from . import data_readers as readers
from . import df_creators
from .data_readers import Reader
from .df_creators import DataFrameCreator
from .df_manipulators import DataFrameManipulator
from .transformer_handler import TransformerHandler


class HistoDataset(torch.utils.data.Dataset):
    """
    TODO
    """

    def __init__(
        self,
        df_creator: U[str, DataFrameCreator],
        path_to_dataset: str,
        df_manipulators: O[U[Sequence[DataFrameManipulator], DataFrameManipulator]] = None,
        data_readers: O[U[str, Reader, Sequence[U[str, Reader]]]] = None,
        feature_readers: O[U[str, Reader, Sequence[U[str, Reader]]]] = None,
        pre_transfs: O[U[TransformerHandler, Callable, Sequence[Callable]]] = None,
        da_transfs: O[U[TransformerHandler, Callable, Sequence[Callable]]] = None,
        seed: int = None,
        return_index: bool = False,
        return_pytorch_tensor: bool = True,
    ):
        """
        TODO

        Arguments:
            transfs: TODO
            access_orders: TODO
            trais_callable_with_batchesnsfs: TODO
            seed: TODO
        """

        super().__init__()

        # save variables
        self.path_to_dataset = path_to_dataset
        self.return_index = return_index
        self.seed = seed
        self.seed_increment = 0
        self.return_pytorch_tensor = return_pytorch_tensor

        # create a DataFrameCreator if str was given
        if isinstance(df_creator, str):
            df_creator = df_creators.CreateDFFromCSV(df_creator)

        # convert or create list of data_readers
        self.data_readers = helper.create_dict_from_variable(data_readers, "data")
        for name in self.data_readers:
            if isinstance(self.data_readers[name], str):
                self.data_readers[name] = readers.ReadFromImageFile(self.data_readers[name])

        # convert or create list of feature_readers
        self.feature_readers = helper.create_dict_from_variable(feature_readers, "feature")
        for name in self.feature_readers:
            if isinstance(self.feature_readers[name], str):
                self.feature_readers[name] = readers.ReadValueFromCSV(self.feature_readers[name])

        # convert or create list of pre_transformation-handler
        if isinstance(pre_transfs, TransformerHandler):
            self.pre_transfs = pre_transfs
        else:
            if self.seed is None:
                self.pre_transfs = TransformerHandler(
                    pre_transfs, seed=random.randint(0, 99999999999)
                )
            else:
                self.pre_transfs = TransformerHandler(pre_transfs, seed=self.seed * int(1e3))

        # convert or create list of da_transformation-handler
        if isinstance(da_transfs, TransformerHandler):
            self.da_transfs = da_transfs
        else:
            self.da_transfs = TransformerHandler(da_transfs)

        # read csv file
        self.df = df_creator(self.path_to_dataset)

        # manipulate the data frame
        if df_manipulators:
            main_data_reader = self.data_readers[list(self.data_readers)[0]]
            if isinstance(df_manipulators, list):
                for df_mani in df_manipulators:
                    self.df = df_mani(self.df, self.path_to_dataset, main_data_reader)
            else:
                self.df = df_manipulators(self.df, self.path_to_dataset, main_data_reader)

        # make sure that the indizes starts by 0
        self.df.reset_index(inplace=True, drop=True)

        # create hash for each row
        self.df["__hash_of_row__"] = self.df.apply(lambda row: int(hash(str(row.values)) % 1e17), 1)

        # for preloading the images / labels
        self.preloaded_data = None
        self.preloaded_features = None
        
    def preload(self, batch_size: int = 128):
        """
        TODO

        Arguments:
            batch_size: TODO

        """

        self.preloaded_data = {}
        self.preloaded_features = {}
        for i in range(0, len(self.df), batch_size):
            data = self.__load_data__(self.df.iloc[i : i + batch_size].index)
            features = self.__load_features__(self.df.iloc[i : i + batch_size].index)

            for k in data:
                if k in self.preloaded_data:
                    self.preloaded_data[k].extend(data[k])
                else:
                    self.preloaded_data[k] = data[k]

            for k in features:
                if k in self.preloaded_features:
                    self.preloaded_features[k].extend(features[k])
                else:
                    self.preloaded_features[k] = features[k]

    def __load_data__(self, idx):
        """
        TODO

        Arguments:
            idx: TODO

        Returns:
            TODO: TODO
        """

        # get asked row
        row = self.df.iloc[idx]

        # load data
        data = {}
        for key, img_loader in self.data_readers.items():
            loaded = img_loader(row, self.path_to_dataset)
            if isinstance(loaded, dict):
                for k in loaded:
                    data[key + "_" + k] = loaded[k]
            else:
                data[key] = loaded

        # pre transform
        data = self.pre_transfs(data, self.__get_hash__(idx))

        return data

    def save(self, path_to_save):
        # TODO-Parameter
        mode = "tiff"
        if mode == "tiff":
            if os.path.exists(path_to_save) == False:
                os.makedirs(path_to_save)

            # check the output of the first sample
            dataset = self[0]
            img_keys = []
            feat_keys = []
            for key in list(dataset):
                if len(dataset[key].shape) in [2, 3]:
                    if os.path.exists(os.path.join(path_to_save, key)) == False:
                        os.mkdir(os.path.join(path_to_save, key))
                    img_keys.append(key)
                else:
                    feat_keys.append(key)

            csv_values = []
            for i in range(len(self)):
                dataset = self[i]
                for img_key in img_keys:
                    path_to_img = os.path.join(path_to_save, img_key)
                    path_to_img = os.path.join(path_to_img, str(i) + ".tiff")
                    im = torchvision.transforms.ToPILImage()(dataset[img_key])
                    im.save(path_to_img)
                csv_row_values = []
                csv_row_values.append(str(i))
                csv_row_values.extend(list(self.df.iloc[i]))
                for feat_key in feat_keys:
                    csv_row_values.append(dataset[feat_key].numpy())
                csv_values.append(csv_row_values)

            with open(os.path.join(path_to_save, "features.csv"), "w", newline="") as csvfile:
                spamwriter = csv.writer(csvfile, delimiter=",")
                spamwriter.writerow(["id"] + list(self.df) + feat_keys)
                for row in csv_values:
                    spamwriter.writerow(row)

        else:
            raise NotImplementedError("")

    def __load_features__(self, idx):
        """
        TODO

        Arguments:
            idx: TODO

        Returns:
            TODO: TODO
        """

        # get asked row
        row = self.df.iloc[idx]

        features = {
            key: feature_loader(row, self.path_to_dataset)
            for key, feature_loader in self.feature_readers.items()
        }

        return features

    def __len__(self) -> int:
        """
        Returns length of DataFrame
        """

        return len(self.df)

    def __get_hash__(self, idx):
        """
        TODO

        Arguments:
            idx: TODO

        Returns:
            TODO: TODO
        """

        # get hashes
        row = self.df.iloc[idx]
        hashes = row["__hash_of_row__"]
        if isinstance(hashes, pd.Series):
            hashes = hashes.values.tolist()
        return hashes

    def __convert_to_tensor__(self, data, idx):
        """
        TODO

        Arguments:
            data: TODO
            idx: TODO

        Returns:
            TODO: TODO
        """

        for key in data:
            if isinstance(idx, list):
                if isinstance(data[key], list):
                    if not torch.is_tensor(data[key][0]):
                        for i, d in enumerate(data[key]):
                            if isinstance(d, np.ndarray):
                                data[key][i] = torch.as_tensor(d.copy())
                            else:
                                data[key][i] = torch.as_tensor(d)
                    data[key] = torch.stack(data[key])
                elif not torch.is_tensor(data[key]):
                    data[key] = torch.as_tensor(data[key])
            elif not torch.is_tensor(data[key]):
                if isinstance(data[key], np.ndarray):
                    data[key] = torch.as_tensor(data[key].copy())
                else:
                    data[key] = torch.as_tensor(data[key])
        return data

    def get_features(self, idx):
        """
        TODO

        Arguments:
            idx: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(idx, np.ndarray):
            idx = list(idx)

        if self.preloaded_data is None:
            features = self.__load_features__(idx)
        else:
            features = {}
            for k in self.preloaded_features:
                if isinstance(idx, list):
                    features[k] = [self.preloaded_features[k][_idx] for _idx in idx]
                else:
                    features[k] = self.preloaded_features[k][idx]
        if self.return_pytorch_tensor:
            features = self.__convert_to_tensor__(features, idx)

        return features

    def get_feature_for_all_rows(self):
        """
        TODO

        Returns:
            TODO: TODO
        """

        if self.preloaded_data:
            return self.preloaded_features
        else:
            return self.get_features(np.arange(len(self)))

    def __getitem__(self, idx) -> dict:
        """
        TODO

        Arguments:
            idx: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(idx, np.ndarray):
            idx = list(idx)

        if self.seed:
            # save random state to reactivate this state after the call
            saved_seed_state = helper.get_random_seed()
            # set seets
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is None:
                worker_id = 0
                num_workers = 1
            else:
                worker_id = worker_info.id
                num_workers = worker_info.num_workers
            seed_to_set = (1 + self.seed_increment + worker_id) * int(1e7) + self.seed

            helper.set_random_seed_with_int(seed_to_set)
            # increase seed increment to use a new random seed at next call
            self.seed_increment += num_workers

        if self.preloaded_data is None:
            data = self.__load_data__(idx)
            features = self.__load_features__(idx)
        else:
            data = {}
            for k in self.preloaded_data:
                if isinstance(idx, list):
                    data[k] = [self.preloaded_data[k][_idx] for _idx in idx]
                else:
                    data[k] = self.preloaded_data[k][idx]
            features = {}
            for k in self.preloaded_features:
                if isinstance(idx, list):
                    features[k] = [self.preloaded_features[k][_idx] for _idx in idx]
                else:
                    features[k] = self.preloaded_features[k][idx]

        # da transforms
        data = self.da_transfs(data, self.__get_hash__(idx))

        if self.seed:
            # reactivate random state
            helper.set_random_seed(*saved_seed_state)

        if self.return_pytorch_tensor:
            data = self.__convert_to_tensor__(data, idx)
            features = self.__convert_to_tensor__(features, idx)

        dic = {}
        dic.update(data)
        dic.update(features)
        if self.return_index:
            dic["idx"] = idx
        return dic



from typing import Sequence

import numpy as np
import torch
import torch.utils


class SamplerBalanced(torch.utils.data.Sampler[int]):
    r"""Samples elements class balanced. If shuffle is False, this sampler works like
    undersampling. But if shuffle True, the elements get shuffled each epoch.

    Arguments:
        data_labels: the labels of each element of the dataset.
        batch_size: the number of ids that this sampler should return in one call.
        shuffle: If true, the elements get shuffled each epoch.
    """

    def __init__(
        self, data_labels: Sequence[int], batch_size: int = 64, shuffle: bool = True
    ) -> None:
        super().__init__(data_source=None)

        self.data_labels = data_labels
        self.batch_size = batch_size
        self.shuffle = shuffle

    @property
    def num_samples(self) -> int:
        return len(self.data_labels)

    def __iter__(self):
        number_of_classes = np.max(self.data_labels) + 1
        indices = []
        num_of_samples_each_class = []
        for class_label in range(number_of_classes):
            indices_of_class = np.argwhere(self.data_labels == class_label)[:, 0]
            if self.shuffle:
                np.random.shuffle(indices_of_class)
            indices.append(indices_of_class)
            num_of_samples_each_class.append(len(indices_of_class))

        num_of_samples = np.min(num_of_samples_each_class)

        bz_each_label = self.batch_size // number_of_classes

        num_of_batches = num_of_samples // bz_each_label

        for i in range(num_of_batches):
            indices_to_return = []
            for class_label in range(number_of_classes):
                indices_to_return.extend(
                    indices[class_label][i * bz_each_label : (i + 1) * bz_each_label]
                )
            yield indices_to_return

    def __len__(self):
        return len(self.data_labels)  # // self.batch_size



from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

from ..helper import helper
from . import transformer_adapters


def __set_values_to_input_dictionary__(dictionary, values):
    input_dic = {}
    for k in dictionary:
        input_dic[k] = values[dictionary[k]]
    return input_dic


def __update_output_dictionary_to_values__(dictionary, values, output_values):
    for k_result, k_dictionary in zip(output_values, dictionary):
        values[dictionary[k_dictionary]] = output_values[k_result]
    return values


class TransformerHandler:
    """
    TransformerHandler administrates the different passed transformations for HistoDataset.
    """
    def __init__(
        self,
        transfs: O[U[Callable, Sequence[Callable]]] = None,
        access_orders: O[Sequence[Sequence[str]]] = None,
        is_callable_with_batches: O[Sequence[bool]] = None,
        seed: O[int] = None,
    ):
        """
        TODO

        Arguments:
            transfs: TODO
            access_orders: TODO
            trais_callable_with_batchesnsfs: TODO
            seed: TODO
        """

        # convert or create list of pre_transformation_adapters
        self.transfs = helper.create_list_from_variable(transfs)

        # create a new access_orders list
        if access_orders:
            if len(self.transfs) != len(access_orders):
                assert AssertionError(
                    "The list of transformers and the list of access_order must be have "
                    + "the same length."
                )
            self.access_orders = access_orders
            for idx, v in enumerate(self.access_orders):
                self.access_orders[idx] = helper.create_list_from_variable(v)
        else:
            self.access_orders = [[] for _ in self.transfs]

        # create a new is_callable_with_batches list
        if is_callable_with_batches:
            if len(self.transfs) != len(is_callable_with_batches):
                assert AssertionError(
                    "The list of transformers and the list of is_callable_with_batches "
                    + "must be have the same length."
                )
            self.is_callable_with_batches = is_callable_with_batches
        else:
            self.is_callable_with_batches = [False for _ in self.transfs]

        self.seed = seed

    def append(self, trans, access_order = None, is_callable_with_batches: bool = False):
        """
        TODO

        Arguments:
            trans: TODO
            access_order: TODO
            is_callable_with_batches: TODO
            
        """

        self.transfs.append(trans)
        if isinstance(access_order, str):
            access_order = [access_order]
        self.access_orders.append(access_order)
        if issubclass(type(trans), transformer_adapters.TransformerAdapter):
            self.is_callable_with_batches.append(trans.is_callable_with_batches)
        else:
            self.is_callable_with_batches.append(is_callable_with_batches)

    def replace(self, trans, access_order=None, is_callable_with_batches: bool = False, index: int = -1):
        """
        TODO

        Arguments:
            trans: TODO
            access_order: TODO
            is_callable_with_batches: TODO
            index: TODO
            
        """

        self.transfs[index] = trans
        if isinstance(access_order, str):
            access_order = [access_order]
        self.access_orders[index] = access_order
        if issubclass(type(trans), transformer_adapters.TransformerAdapter):
            self.is_callable_with_batches[index] = trans.is_callable_with_batches
        else:
            self.is_callable_with_batches[index] = is_callable_with_batches

    def __call_single_transform_with_dict__(
        self, item_or_batch, hashes, transf, dictionary, batch_callable
    ):
        """
        TODO

        Arguments:
            item_or_batch: TODO
            hashes: TODO
            transf: TODO
            dictionary: TODO
            batch_callable: TODO

        Returns:
            TODO: TODO
        """

        is_batch = isinstance(hashes, list)

        if is_batch is False:
            if self.seed:
                helper.set_random_seed_with_int(int(self.seed + hashes))
            input_dic = __set_values_to_input_dictionary__(dictionary, item_or_batch)
            result = transf(**input_dic)
            return __update_output_dictionary_to_values__(dictionary, item_or_batch, result)
        elif batch_callable and self.seed:
            input_dic = __set_values_to_input_dictionary__(dictionary, item_or_batch)
            result = transf(**input_dic)
            item_or_batch.update(result)
            return item_or_batch
        else:
            for i in range(len(item_or_batch[list(item_or_batch)[0]])):
                item = {k: item_or_batch[k][i] for k in item_or_batch}
                item = self.__call_single_transform_with_dict__(
                    item, hashes[i], transf, dictionary, batch_callable
                )
                for k in item:
                    item_or_batch[k][i] = item[k]
            return item_or_batch

    def __call_single_transform_with_callable__(
        self, item_or_batch, hashes, transf, order, batch_callable
    ):
        """
        TODO

        Arguments:
            item_or_batch: TODO
            hashes: TODO
            transf: TODO
            order: TODO
            batch_callable: TODO

        Returns:
            TODO: TODO
        """

        is_batch = isinstance(hashes, list)

        if order is None or order == [] or order[0] is None or order[0] == "":
            order = list(item_or_batch)

        tmp_saved_seed_state = helper.get_random_seed()

        for i, name in enumerate(order):

            element = item_or_batch.pop(name)

            if is_batch is False:
                if self.seed:
                    helper.set_random_seed_with_int(int(self.seed + hashes))
                result = transf(element)
            elif batch_callable and self.seed is None:
                result = transf(element)
            else:
                result = {}
                for j, item in enumerate(element):
                    if self.seed:
                        helper.set_random_seed_with_int(int(self.seed + hashes[j]))

                    result_item = transf(item)

                    if isinstance(result_item, dict):
                        if j == 0:
                            for k in result_item:
                                result[name + "_" + k] = [result_item[k]]
                        else:
                            for k in result_item:
                                result[name + "_" + k].append(result_item[k])
                    else:
                        if j == 0:
                            result[name] = [result_item]
                        else:
                            result[name].append(result_item)

            if isinstance(result, dict):
                for k in result:
                    if k.startswith(name):
                        item_or_batch[k] = result[k]
                    else:
                        item_or_batch[name + "_" + k] = result[k]
            else:
                item_or_batch[name] = result

            # use the same seed for each data that contains to the same image.
            if i < len(order) - 1:
                helper.set_random_seed(*tmp_saved_seed_state)

        return item_or_batch

    def __call__(self, item_or_batch, hashes):
        """
        TODO

        Arguments:
            item_or_batch: TODO
            hashes: TODO

        Returns:
            TODO: TODO
        """

        if self.seed:
            saved_seed_state = helper.get_random_seed()

        for (transf, order, batch_callable) in zip(
            self.transfs, self.access_orders, self.is_callable_with_batches
        ):
            if isinstance(order, dict):
                item_or_batch = self.__call_single_transform_with_dict__(
                    item_or_batch, hashes, transf, order, batch_callable
                )
            else:
                item_or_batch = self.__call_single_transform_with_callable__(
                    item_or_batch, hashes, transf, order, batch_callable
                )

        if self.seed:
            helper.set_random_seed(*saved_seed_state)

        return item_or_batch

    def __len__(self):
        """
        Returns length of attribute transfs.
        """
        
        return len(self.transfs)

    def __call_with_prints__(self, item_or_batch, hashes):        
        """
        TODO

        Arguments:
            item_or_batch: TODO
            hashes: TODO

        Returns:
            TODO: TODO
        """

        if self.seed:
            saved_seed_state = helper.get_random_seed()

        index = 0
        for (transf, order, batch_callable) in zip(
            self.transfs, self.access_orders, self.is_callable_with_batches
        ):
            print(index, type(self.transfs))
            print("All Input-keys:", list(item_or_batch))
            print("Access Order", self.access_orders)
            index += 1

            if isinstance(order, dict):
                item_or_batch = self.__call_single_transform_with_dict__(
                    item_or_batch, hashes, transf, order, batch_callable
                )
            else:
                item_or_batch = self.__call_single_transform_with_callable__(
                    item_or_batch, hashes, transf, order, batch_callable
                )
            print("All Output-keys:", list(item_or_batch))
            print()

        if self.seed:
            helper.set_random_seed(*saved_seed_state)

        return item_or_batch



import random
from enum import IntEnum

import numpy as np
import torch


class DataLoaderMode(IntEnum):
    TRAINING = 1
    VALIDATION = 2
    TRAINING_AND_VALIDATION = 3
    TEST = 4
    ALL = 7


def collate_fn(
    batch,
):  # Here we would need two collate function for train and validation!
    r"""
    This collate_fn function can be used with the "SamplerSimple" and "SamplerBalanced".
    The batch size should be set in the Sampler and not in the DataLoader.
    """
    data = batch[0][0]
    target = torch.as_tensor(batch[0][1], dtype=torch.int64)

    return data, target


# EIGENE FUNKTIOn, bekommt ein DATASET und splitted diese in zwei oder drei neue Datasets
# mode: DataLoaderMode,
# cross_validation_fold: O[U[Sequence[int], int]] = None,


def create_circular_mask(h, w, center=None, radius=None):

    if center is None:  # use the middle of the image
        center = (int(w / 2), int(h / 2))
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[1], center[0], w - center[1], h - center[0])

    y, x = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((y - center[1]) ** 2 + (x - center[0]) ** 2)

    mask = dist_from_center <= radius
    return mask


def create_list_from_variable(variable):
    if variable is None:
        return []
    elif isinstance(variable, list):
        return variable
    else:
        return [variable]


def create_dict_from_variable(variable, default_name):
    if variable is None:
        return {}
    elif isinstance(variable, list):
        dic = {}
        for idx, element in enumerate(variable):
            dic[str(default_name) + "_" + str(idx)] = element
        return dic
    elif isinstance(variable, dict):
        return variable
    else:
        return {default_name: variable}


def get_random_seed():
    module_random = random.getstate()
    module_numpy = np.random.get_state()
    module_torch = torch.get_rng_state()
    return module_random, module_numpy, module_torch


def set_random_seed(module_random, module_numpy, module_torch):
    random.setstate(module_random)
    np.random.set_state(module_numpy)
    torch.set_rng_state(module_torch)


def set_random_seed_with_int(integer):
    random.seed(integer)
    np.random.seed(random.randint(0, 99999999))
    torch.random.manual_seed(random.randint(0, 99999999))


def read_from_row(row, expression):
    return expression.format(**row)


def labels_to_weights(labels):
    labels = np.array(labels)
    label_prob = np.zeros_like(labels, dtype=np.float)
    for l in np.unique(labels):
        prob = sum(labels == l) / float(len(labels))
        label_prob[labels == l] = 1 - prob
    return label_prob



import math
import random

import matplotlib.pyplot as plt
import numpy as np
import openslide
import pandas as pd
from scipy.ndimage.measurements import center_of_mass
from skimage import morphology
from skimage.color import rgb2hed
from skimage.filters import threshold_otsu
from skimage.segmentation import slic


class DataFrameManipulator:
    """
    Parent class for all manipulators. Subclasses need to implement their call method.
    """

    def __init__(self):
        pass

    def __call__(self, df, path_to_image, reader):
        pass


class BoundingBoxToCenterCoordinates(DataFrameManipulator):
    """
    TODO
    """

    def __init__(
        self, column_of_bbox: str, name_of_x_column: str = "__x__", name_of_y_column: str = "__y__"
    ):
        self.column_of_bbox = column_of_bbox
        self.name_of_x_column = name_of_x_column
        self.name_of_y_column = name_of_y_column

    def __call__(self, df, path_to_image, reader):
        x = []
        y = []
        for _, row in df.iterrows():
            x.append((row[self.column_of_bbox][1] + row[self.column_of_bbox][3]) / 2)
            y.append((row[self.column_of_bbox][0] + row[self.column_of_bbox][2]) / 2)
        df[self.name_of_x_column] = x
        df[self.name_of_y_column] = y
        return df


class EditDFImageFilter(DataFrameManipulator):
    """
    TODO
    """
    def __init__(self, min_std):
        super().__init__()
        
        
    def __call__(self, df, path_to_image: str, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: TODO
            reader: TODO

        Returns:
            TODO: TODO
        """
        import matplotlib.pyplot as plt
        for index, row in df.iterrows():
            img = reader(row, path_to_image)
            img = img[int(img.shape[0]*0.25):int(img.shape[0]*0.75),int(img.shape[1]*0.25):int(img.shape[1]*0.75)]
            img_max_red_blue = (np.array(img[:,:,0],dtype=float)<235) | (np.array(img[:,:,1],dtype=float)<150)
            max_red_blue = np.sum(img_max_red_blue) / (img.shape[0] * img.shape[1])
            if max_red_blue < 0.25:
                df.drop(index, inplace=True)
        return df
    
class EditDFImageGrid(DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, return_image_size, perc_overlapping: float = 0.0, input_im_size=None):
        """
        TODO

        Arguments:
            return_image_size: TODO
            perc_overlapping: TODO
            input_im_size: TODO
        """

        super().__init__()
        self.return_image_size = return_image_size
        self.perc_overlapping = perc_overlapping
        self.input_im_size = input_im_size

    def get_points(self, im_size1d):
        """
        TODO

        Arguments:
            im_size1d: TODO

        """

        im_size_center = im_size1d - self.return_image_size
        shift = self.return_image_size * (1 - self.perc_overlapping)
        locations = np.arange(0, im_size_center + 0.0001, shift)

        return np.array(locations + (im_size1d - locations[-1]) / 2, dtype=int)
    
    def __call__(self, df, path_to_image: str, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: TODO
            reader: TODO

        Returns:
            TODO: TODO
        """

        if self.input_im_size:
            input_im_size = self.input_im_size

        rows = {name: [] for name in list(df)}
        rows["__x__"] = []
        rows["__y__"] = []
        for _, row in df.iterrows():
            if self.input_im_size is None:
                input_im_size = reader.get_image_size(row, path_to_image)
            xs = self.get_points(input_im_size[0])
            ys = self.get_points(input_im_size[1])

            for x in xs:
                for y in ys:
                    row = dict(row)
                    row["__x__"] = x
                    row["__y__"] = y

                    for k in row:
                        rows[k].append(row[k])

        return pd.DataFrame(rows)


def get_mask(rgb):
    """
    TODO

    Arguments:
        rgb: TODO

    Returns:
        TODO: TODO
    """

    rgb = np.array(rgb, dtype=float) / 255.0

    ihc_hed = rgb2hed(rgb)

    gray = ihc_hed[..., 0] + ihc_hed[..., 1]  # - ihc_hed[..., 2]

    counts, positions = np.histogram(gray.flatten(), bins=50, range=(0.001, 0.04))
    argmax = np.argmax(counts)

    threshold1 = None
    threshold2 = None

    for pos in range(argmax, 48):
        if threshold1 is None:
            if counts[pos + 1] > counts[pos] and counts[pos + 2] > counts[pos + 1]:
                threshold1 = positions[pos]
        if threshold2 is None:
            if counts[pos + 1] > counts[pos]:
                threshold2 = positions[pos]
    if threshold1:
        threshold = threshold1
    elif threshold2:
        threshold = threshold2
    else:
        threshold = 0.04

    # use otsu as second threshold
    values_for_otsu = gray.flatten()
    values_for_otsu = values_for_otsu[values_for_otsu > 0.001]
    values_for_otsu = values_for_otsu[values_for_otsu < 0.04]
    threshold_from_otsu = threshold_otsu(values_for_otsu)

    # use the smaller threshold
    mask = gray > min([threshold, threshold_from_otsu])

    mask = morphology.binary_opening(mask, morphology.disk(1))
    mask = morphology.binary_closing(mask, morphology.disk(1))
    mask = morphology.binary_opening(mask, morphology.disk(3))
    mask = morphology.binary_closing(mask, morphology.disk(9))

    return mask


class EditDFWSIGrid(DataFrameManipulator):
    """
    TODO
    """

    def __init__(
        self,
        return_image_size,
        perc_overlapping: float = 0.0,
        plt_coordinates: bool = False,
        fn_thumbnail_to_mask=get_mask,
    ):
        """
        TODO

        Arguments:
            return_image_size: TODO
            perc_overlapping: TODO
            plt_coordinates: TODO
            fn_thumbnail_to_mask: TODO

        """
        super().__init__()
        self.return_image_size = return_image_size
        self.perc_overlapping = perc_overlapping
        self.plt_coordinates = plt_coordinates
        self.fn_thumbnail_to_mask = fn_thumbnail_to_mask

    def __get_coordinates__(self, thumbnail, thumbnail_mpp, return_mpp, min_mpp):
        """
        TODO

        Arguments:
            thumbnail: TODO
            thumbnail_mpp: TODO
            return_mpp: TODO
            min_mpp: TODO

        Returns:
            TODO: TODO
        """

        return_mpp = float(return_mpp)
        thumbnail_mpp = float(thumbnail_mpp)
        # see: https://colab.research.google.com/github/TIA-Lab
        # /tiatoolbox/blob/master/examples/example_wsiread.ipynb#scrollTo=MZ_yqoGJ_-6i
        wsi_thumb_mask = self.fn_thumbnail_to_mask(thumbnail)

        return_image_stride = self.return_image_size * (1 - self.perc_overlapping)

        lores_patch_size = int(return_image_stride / (thumbnail_mpp / return_mpp))
        nr_expected_rois = math.ceil(np.sum(wsi_thumb_mask) / (lores_patch_size ** 2))

        print(f"Created ~{nr_expected_rois} patches.")

        wsi_rois_mask = slic(
            thumbnail, mask=wsi_thumb_mask, n_segments=nr_expected_rois, compactness=1000, sigma=1
        )

        lores_rois_center = center_of_mass(
            wsi_rois_mask, labels=wsi_rois_mask, index=np.unique(wsi_rois_mask)[1:]
        )
        lores_rois_center = np.array(lores_rois_center)  # coordinates is Y, X
        lores_rois_center = lores_rois_center.astype(np.int32)
        selected_indices = wsi_thumb_mask[lores_rois_center[:, 0], lores_rois_center[:, 1]]
        lores_rois_center = lores_rois_center[selected_indices]

        distance_to_boarder = []
        for center_x, center_y in zip(lores_rois_center[:, 1], lores_rois_center[:, 0]):
            y, x = np.ogrid[: wsi_thumb_mask.shape[0], : wsi_thumb_mask.shape[1]]
            dist_from_center = np.sqrt((y - center_y) ** 2 + (x - center_x) ** 2)
            dist = np.min(dist_from_center[~wsi_thumb_mask])
            distance_to_boarder.append(dist)
        distance_to_boarder = np.array(distance_to_boarder)

        if self.plt_coordinates:
            # show the patches region and their centres of mass
            plt.figure(figsize=(20, 6))
            plt.subplot(1, 4, 1)
            # plt.subfigure(subf)
            plt.imshow(thumbnail)
            plt.subplot(1, 4, 2)
            plt.imshow(thumbnail)
            plt.scatter(lores_rois_center[:, 1], lores_rois_center[:, 0], s=5)
            plt.scatter(lores_rois_center[0:1, 1], lores_rois_center[0:1, 0], s=18, c="g")
            plt.scatter(lores_rois_center[1:2, 1], lores_rois_center[1:2, 0], s=18, c="y")
            plt.axis("off")
            plt.subplot(1, 4, 3)
            plt.scatter(
                lores_rois_center[:, 1], -lores_rois_center[:, 0], c=distance_to_boarder, s=5
            )
            plt.axis("off")
            plt.subplot(1, 4, 4)
            plt.imshow(wsi_thumb_mask)
            plt.show()

        return lores_rois_center * (thumbnail_mpp / min_mpp), distance_to_boarder * (
            thumbnail_mpp / min_mpp
        )

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """

        rows = {name: [] for name in list(df)}
        rows["__x__"] = []
        rows["__y__"] = []
        rows["__dist__"] = []
        for _, row in df.iterrows():
            try:
                thumbnail, min_mpp = reader.read_thumbnail(
                    row, path_to_image, 20, return_min_mpp=True
                )
            except openslide.OpenSlideUnsupportedFormatError:
                print("Warning: The file is damaged or has missing files.")
                continue
            coordinates, distance_to_boarders = self.__get_coordinates__(
                thumbnail, 20, reader.mpp, min_mpp[0]
            )

            for x, y, dist in zip(coordinates[:, 1], coordinates[:, 0], distance_to_boarders):
                row = dict(row)
                row["__x__"] = int(x)
                row["__y__"] = int(y)
                row["__dist__"] = dist

                for k in row:
                    rows[k].append(row[k])

        return pd.DataFrame(rows)

    
class RandomDataSetSplitFilter(DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, num_of_folds, folds_to_use, seed=None, name_of_fold_column='__folds__'):
        """
        TODO

        Arguments:
            percentage_to_filter: TODO
            mode: TODO
            seed: TODO

        """
        DataFrameManipulator.__init__(self)
        self.num_of_folds = num_of_folds
        self.folds_to_use = folds_to_use
        self.seed = seed
        self.name_of_fold_column = name_of_fold_column

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        if self.name_of_fold_column not in df:
            rng = np.random.default_rng(self.seed)
            df[self.name_of_fold_column] = rng.integers(self.num_of_folds, size=len(df))
        
        return df[df[self.name_of_fold_column].isin(self.folds_to_use)]
    

class RandomDataSetSplitFilterByColumnValue(DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, column, num_of_folds, folds_to_use, seed=None, name_of_fold_column='__folds__'):
        """
        TODO

        Arguments:
            percentage_to_filter: TODO
            mode: TODO
            seed: TODO

        """
        DataFrameManipulator.__init__(self)
        self.column = column
        self.num_of_folds = num_of_folds
        self.folds_to_use = folds_to_use
        self.seed = seed
        self.name_of_fold_column = name_of_fold_column

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        if self.name_of_fold_column not in df:
            rng = np.random.default_rng(self.seed)
            g = df.groupby(self.column)
            random_numbers = rng.integers(self.num_of_folds, size=len(g.indices.keys()))
            folds = np.zeros(len(df), dtype=int)
            for k, r in zip(g.indices, random_numbers):
                folds[g.indices[k]] = r
            df[self.name_of_fold_column] = folds
        
        return df[df[self.name_of_fold_column].isin(self.folds_to_use)]
    
    
class DataFrameFilter:
    """
    TODO
    """

    def __init__(self, mode):
        """
        TODO

        Arguments:
            mode: TODO

        """
        self.mode = mode

    def __call__(self, df, indices):
        """
        TODO

        Arguments:
            df: TODO
            indices: TODO

        Returns:
            TODO: TODO
        """

        if self.mode == 0:
            return df.loc[indices]
        elif self.mode == 1:
            return df.loc[~df.index.isin(indices)]
        else:
            return df.loc[indices], df.loc[~df.index.isin(indices)]



        
class HematoxylinFilter(DataFrameFilter, DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, image_size, mode=0):
        """
        TODO

        Arguments:
            percentage_to_filter: TODO
            mode: TODO
            seed: TODO

        """
        DataFrameFilter.__init__(self, mode)
        DataFrameManipulator.__init__(self)
        from skimage.color import rgb2hed, hed2rgb
        self.center_start = image_size - (image_size/(2**0.5))
        
    def hematoxylin_filter(self, patch):
        ihc_hed = rgb2hed(patch[128:512-128,128:512-128])
        return np.mean(ihc_hed[:, :, 0] > 0.05) > 0.03
    
    def __call__(self, df, path_to_dataset, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        indices = []
        for index, row in df.iterrows():
            loaded = reader(row, path_to_dataset)
            if self.hematoxylin_filter(loaded):
                indices.append(index)
        return DataFrameFilter.__call__(self, df, indices)
        

class RandomFilter(DataFrameFilter, DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, percentage_to_filter, mode=0, seed=None):
        """
        TODO

        Arguments:
            percentage_to_filter: TODO
            mode: TODO
            seed: TODO

        """

        DataFrameFilter.__init__(self, mode)
        DataFrameManipulator.__init__(self)
        self.percentage_to_filter = percentage_to_filter
        self.seed = seed

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """

        r = random.Random(self.seed)
        indices = r.sample(list(df.index), int(len(df) * self.percentage_to_filter))
        return DataFrameFilter.__call__(self, df, indices)


class RandomFilterByColumnValue(DataFrameFilter, DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, column, percentage_to_filter, mode=0, seed=None):
        """
        TODO

        Arguments:
            column: TODO
            percentage_to_filter: TODO
            mode: TODO
            seed: TODO

        """

        DataFrameFilter.__init__(self, mode)
        DataFrameManipulator.__init__(self)
        self.column = column
        self.percentage_to_filter = percentage_to_filter
        self.seed = seed

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        uv = list(df[self.column].unique())
        r = random.Random(self.seed)
        choosen = r.sample(uv, int(len(uv) * self.percentage_to_filter))
        indices = list(df[df[self.column].isin(choosen)].index)
        return DataFrameFilter.__call__(self, df, indices)


class LambdaRowFilter(DataFrameFilter, DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, lambda_function, mode=0):
        """
        TODO

        Arguments:
            lambda_function: TODO
            mode: TODO

        """

        DataFrameFilter.__init__(self, mode)
        DataFrameManipulator.__init__(self)
        self.lambda_function = lambda_function

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        return DataFrameFilter.__call__(self, df, df[df.apply(self.lambda_function, 1)].index)


class LambdaDFFilter(DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, lambda_function):
        """
        TODO

        Arguments:
            lambda_function: TODO
            mode: TODO

        """
        DataFrameManipulator.__init__(self)
        self.lambda_function = lambda_function

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """
        return self.lambda_function(df)
    
    
class ColumnFilter(DataFrameFilter, DataFrameManipulator):
    """
    TODO
    """

    def __init__(self, name_of_column, allowed_values, mode=0):
        """
        TODO

        Arguments:
            name_of_column: TODO
            allowed_values: TODO
            mode: TODO

        """

        DataFrameFilter.__init__(self, mode)
        DataFrameManipulator.__init__(self)
        self.name_of_column = name_of_column
        self.allowed_values = allowed_values

    def __call__(self, df, path_to_image, reader):
        """
        TODO

        Arguments:
            df: TODO
            path_to_image: Absolute path to image.
            reader: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(self.allowed_values, list):
            return DataFrameFilter.__call__(
                self, df, df[df[self.name_of_column].isin(self.allowed_values)].index
            )
        else:
            return DataFrameFilter.__call__(
                self, df, df[df[self.name_of_column] == self.allowed_values].index
            )



import os
from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

import numpy as np

from ..base import data_readers, df_creators, df_manipulators
from ..base.histo_dataset import HistoDataset


def midog_classification_patches(
    root: str,  # '/data/ldap/histopathologic/original_read_only/MIDOG'
    patch_size: int,
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: str = "all",
    splitting_seed: int = 1,
    scanner_to_use: O[U[int, list]] = None,
    seed: O[int] = None,
) -> HistoDataset:
    """
    Creates the MIDOG classification dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at the README.md of the git repository (TODO).

    Arguments:
        root: Absolute or relative path to the dataset files.
        transformation: A callable or a list of callable transformations.
        pre_transformation: A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "validation", "test", "all"}; size of subsets: train 3414,val 505,test 516
        splitting_seed: Seed used for splitting the data set
        scanner_to_use: The midog classification dataset includes three scanners.
                        This is a list or a integer that describes which scanner
                        is used. Allowed values are 0,1 or 2.
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position. For example:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
                    If we create the exact same dataset the transformation seeds
                    will be the same:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
        patch_size: The images get cropped from the original image. Here you can
                decide which size arround the mitosis get cropped.

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary in form
            dict['images': torch.Tensor, 'feature': torch.Tensor].
            And the label is between 0 and 1.
            The feature 0 stands for Mitosis, the label 1 stands for
            non-mitotic.
    """
    manipulators_to_use = []
    
    # handle training, validation and test cases
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        folds_to_use = [0,1]
    elif subset.startswith("val"):
        folds_to_use = [2]
    elif subset.startswith("te"):  
        folds_to_use = [3]
    elif subset == "all":
        folds_to_use = [0,1,2,3]
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "validation", "test", "all"].'
        )
    
    """
    # If this is used instead of the following, then each large original image, will be
    # set to one of the subsets with all the datapoints in it.
    manipulators_to_use.append(
        df_manipulators.RandomDataSetSplitFilterByColumnValue(
            "image_id", num_of_folds=4, folds_to_use=folds_to_use, seed=splitting_seed
        )
    )
    """
    manipulators_to_use.append(
        df_manipulators.RandomDataSetSplitFilter(
            num_of_folds=4, folds_to_use=folds_to_use, seed=splitting_seed
        )
    )

    # handle scanner values
    if scanner_to_use:
        if isinstance(scanner_to_use, int):
            allowed_values = np.arange(50) + 1 + 50 * scanner_to_use
        elif isinstance(scanner_to_use, list):
            allowed_values = []
            for s in scanner_to_use:
                allowed_values.extend(np.arange(50) + 1 + 50 * s)
        else:
            raise NotImplementedError(
                'The parameter "scanner_to_use" needs to be None, an integer or a list.'
            )
        manipulators_to_use.append(df_manipulators.ColumnFilter("image_id", allowed_values))

    manipulators_to_use.append(df_manipulators.BoundingBoxToCenterCoordinates("bbox"))

    data_reader = data_readers.ReadFromImageFile(
        "images/{image_id:03d}.tiff", return_image_size=patch_size
    )

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromJSON("MIDOG.json", "annotations"),
        root,
        df_manipulators=manipulators_to_use,
        data_readers=data_reader,
        feature_readers=data_readers.ReadValueFromCSV(r"{category_id}", encoded_values=["1", "2"]),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


""" TODO: how the offline dataset was generated
# MIDOG
from histodata.datasets import midog

from torchvision import transforms

transform = transforms.Compose([
    transforms.ToTensor(),
])
for patch_size in [423, 200]:
    ds = midog.midog_classification_patches(
        '/data/ldap/histopathologic/original_read_only/MIDOG',
        transformation=transform,
        subset='all',
        splitting_seed=1,
        patch_size=patch_size,
    )
    ds.save('/data/ldap/histopathologic/original_read_only/MIDOG_classification_' + str(patch_size))
"""


def midog_classification_patches_preprocessed(
    root: str,  # '/data/ldap/histopathologic/original_read_only/MIDOG2/'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: O[str] = "train",
    seed: O[int] = None,
) -> HistoDataset:
    """
    Creates the midog dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO

    Arguments:
        root: absolute or relative path to the dataset files.
        transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "valid", "test"}; size of subsets: train 3414,val 505,test 516
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position. For example:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
                    If we create the exact same dataset the transformation seeds
                    will be the same:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary in form
            dict['images': torch.Tensor, 'feature': torch.Tensor].
            And the label is between 0 and 1.
            The feature 0 stands for Mitosis, the label 1 stands for
            non-mitotic.
    """
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        df_manipulator = df_manipulators.ColumnFilter(
            name_of_column='__folds__', allowed_values=[0,1]
        )
    elif subset.startswith("val"):
        df_manipulator = df_manipulators.ColumnFilter(
            name_of_column='__folds__', allowed_values=[2]
        )
    elif subset.startswith("te"):
        df_manipulator = df_manipulators.ColumnFilter(
            name_of_column='__folds__', allowed_values=[3]
        )
    elif subset.startswith("all"):
        df_manipulator = None
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "valid", "test"].'
        )

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromCSV("features.csv"),
        root,
        df_manipulators=df_manipulator,
        data_readers="data/{id}.tiff",
        feature_readers=data_readers.ReadValueFromCSV(r"{feature}", encoded_values=["0", "1"]),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="histodata",
    version="0.2.21",
    author="Jonas Annuscheit",
    author_email="Jonas.Annuscheit@htw-berlin.de",
    description="Histopathology datasets for PyTorch environments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.tools.f4.htw-berlin.de/cbmi-charite/histodata",
    packages=setuptools.find_packages(
        exclude=(
            "examples",
            "tests",
        )
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "imagesize==1.2.0",
        "tiatoolbox==0.5.1",
        "torchvision>=0.6.1",
    ],
)
